import argparse
import sys

from yamk import __version__
from yamk.lib import SUPPORTED_FILE_EXTENSIONS
from yamk.make import MakeCommand

sys.tracebacklimit = 0


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        prog="yamk", description="Yet another make command"
    )

    # positional arguments
    parser.add_argument("target", help="the target for yamk")

    # optional arguments
    parser.add_argument(
        "-a",
        "--assume",
        action="append",
        metavar="dependency",
        default=[],
        help="assume that dependency is up to date (works only with phony ones)",
    )
    parser.add_argument(
        "-b",
        "--bare",
        action="store_true",
        help="build only the target, without checking the dependencies",
    )
    parser.add_argument(
        "-c",
        "--cookbook",
        metavar="cookbook",
        help="the path to the cookbook",
    )
    parser.add_argument(
        "-d",
        "--directory",
        metavar="dir",
        default=".",
        help="the path to the directory that contains the cookbook",
    )
    parser.add_argument(
        "-f",
        "--force",
        action="store_true",
        help="rebuild all dependencies and the target",
    )
    parser.add_argument(
        "-n",
        "--dry-run",
        action="store_true",
        help="only print the commands to be executed",
    )
    parser.add_argument(
        "-r",
        "--retries",
        metavar="retries",
        type=int,
        default=0,
        help="retry commands for <retries> number of times",
    )
    parser.add_argument(
        "-s",
        "--shell",
        metavar="shell",
        help="the path to the shell used to execute the commands",
    )
    parser.add_argument(
        "-t",
        "--cookbook-type",
        choices=set(SUPPORTED_FILE_EXTENSIONS.values()),
        help="the type of the cookbook. defaults to file extension",
    )
    parser.add_argument(
        "-T",
        "--time",
        action="store_true",
        help="print a timing report",
    )
    parser.add_argument(
        "-V",
        "--version",
        action="version",
        version=f"%(prog)s {__version__}",
        help="print the version and exit",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help="increase the level of verbosity",
    )
    parser.add_argument(
        "-x",
        "--variable",
        action="append",
        metavar="KEY=value",
        dest="variables",
        default=[],
        help="a list of variables to override the ones set in the cookbook, "
        "which should be in the form <variable>=<value>",
    )

    # extra arguments
    parser.add_argument(
        "extra",
        nargs=argparse.REMAINDER,
        help="extra args to be passed to the recipe",
    )

    return parser.parse_args()


def main() -> None:
    args = parse_args()
    MakeCommand(args).make()



from __future__ import annotations

from functools import reduce
from pathlib import Path
from typing import Any, cast

from yamk.types import Pathlike


class Function:
    name: str

    def __init__(self, base_dir: Path):
        self.base_dir = base_dir

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        raise NotImplementedError(f"{self.__class__.__name__} must be callable")


class Glob(Function):
    name = "glob"

    def __call__(self, pattern: str) -> list[str]:
        return [path.as_posix() for path in self.base_dir.glob(pattern)]


class Sort(Function):
    name = "sort"

    def __call__(self, *args: Any) -> list[Any]:
        return sorted(*args)


class Exists(Function):
    name = "exists"

    def __call__(self, path: Pathlike | list[Pathlike]) -> bool | list[bool]:
        if isinstance(path, list):
            return [cast(bool, self(file)) for file in path]
        return self.base_dir.joinpath(path).exists()


class Name(Function):
    name = "name"

    def __call__(self, path: Pathlike | list[Pathlike]) -> str | list[str]:
        if isinstance(path, list):
            return [cast(str, self(file)) for file in path]
        return self.base_dir.joinpath(path).name


class Stem(Function):
    name = "stem"

    def __call__(self, path: Pathlike | list[Pathlike]) -> str | list[str]:
        if isinstance(path, list):
            return [cast(str, self(file)) for file in path]
        return self.base_dir.joinpath(path).stem


class Suffix(Function):
    name = "suffix"

    def __call__(self, path: Pathlike | list[Pathlike]) -> str | list[str]:
        if isinstance(path, list):
            return [cast(str, self(file)) for file in path]
        return self.base_dir.joinpath(path).suffix


class Parent(Function):
    name = "parent"

    def __call__(self, path: Pathlike | list[Pathlike]) -> str | list[str]:
        if isinstance(path, list):
            return [cast(str, self(file)) for file in path]
        return self.base_dir.joinpath(path).parent.as_posix()


class ChangeSuffix(Function):
    name = "change_suffix"

    def __call__(self, path: Pathlike | list[Pathlike], suffix: str) -> str | list[str]:
        if isinstance(path, list):
            return [cast(str, self(file, suffix)) for file in path]
        return self.base_dir.joinpath(path).with_suffix(suffix).as_posix()


class ChangeParent(Function):
    name = "change_parent"

    def __call__(
        self, path: Pathlike | list[Pathlike], parent: Pathlike
    ) -> str | list[str]:
        if isinstance(path, list):
            return [cast(str, self(file, parent)) for file in path]
        name = self.base_dir.joinpath(path).name
        return self.base_dir.joinpath(parent, name).as_posix()


class PWD(Function):
    name = "pwd"

    def __call__(self) -> str:
        return self.base_dir.as_posix()


class FilterOut(Function):
    name = "filter_out"

    def __call__(self, odd: Any, obj: list[Any]) -> list[Any]:
        return list(filter(lambda x: x != odd, obj))


class TernaryIf(Function):
    name = "ternary_if"

    def __call__(
        self, condition: bool, if_true: Any, if_false: Any  # noqa: FBT001
    ) -> Any:
        return if_true if condition else if_false


class Substitute(Function):
    name = "sub"

    def __call__(self, old: str, new: str, obj: list[str] | str) -> list[str] | str:
        if isinstance(obj, list):
            return [cast(str, self(old, new, string)) for string in obj]
        return obj.replace(old, new)


class Merge(Function):
    name = "merge"

    @staticmethod
    def _as_list(obj: Any | list[Any]) -> list[Any]:
        return obj if isinstance(obj, list) else [obj]

    def __call__(self, *args: Any | list[Any]) -> list[Any]:
        return reduce(lambda x, y: self._as_list(x) + self._as_list(y), args)


functions = {function.name: function for function in Function.__subclasses__()}



from __future__ import annotations

import argparse
import itertools
import pathlib
import re
import subprocess
import sys
import warnings
from time import perf_counter_ns, sleep
from typing import Any, cast

from dj_settings import SettingsParser
from packaging.version import Version

from yamk import __version__, lib


class MakeCommand:
    def __init__(self, args: argparse.Namespace):
        self.verbosity = args.verbose
        if self.verbosity:
            if self.verbosity > 1:
                print(args)
            sys.tracebacklimit = 9999
        self.regex_recipes: dict[str, lib.Recipe] = {}
        self.static_recipes: dict[str, lib.Recipe] = {}
        self.aliases: dict[str, str] = {}
        self.target = args.target
        self.bare = args.bare
        self.force_make = args.force
        self.extra = args.extra
        self.retries = args.retries
        self.dry_run = args.dry_run
        cookbook = self.find_cookbook(args)
        self.base_dir = cookbook.parent
        self.phony_dir = self.base_dir.joinpath(".yamk")
        self.arg_vars = dict(var.split("=", maxsplit=1) for var in args.variables)
        parsed_cookbook = SettingsParser(cookbook, force_type=args.cookbook_type).data
        self.globals = parsed_cookbook.pop("$globals", {})
        self.version = self._get_version()
        if self.version > __version__:
            raise RuntimeError(f"This cookbook requires an yamk >= v{self.version}")
        self.up_to_date = args.assume
        self._parse_recipes(parsed_cookbook)
        self.subprocess_kwargs = {
            "shell": True,
            "cwd": self.base_dir,
            "executable": self.globals.get("shell") or args.shell,
        }
        self.print_timing_report = args.time
        self.reports: list[lib.CommandReport] = []

    @staticmethod
    def find_cookbook(args: argparse.Namespace) -> pathlib.Path:
        absolute_path = pathlib.Path(args.directory).absolute()
        if args.cookbook:
            return absolute_path.joinpath(args.cookbook)

        cookbooks = (
            absolute_path.joinpath("cookbook").with_suffix(suffix)
            for suffix in lib.SUPPORTED_FILE_EXTENSIONS
        )
        for cookbook in cookbooks:
            if cookbook.exists():
                return cookbook
        raise FileNotFoundError(f"No candidate cookbook found in {args.directory}")

    def make(self) -> None:
        dag = self._preprocess_target()
        for node in filter(lambda x: x.should_build, dag):
            self._make_target(cast(lib.Recipe, node.recipe))
        if self.print_timing_report:
            lib.print_reports(self.reports)

    def _run_command(self, command: str) -> int:
        status = 0
        if self.dry_run:
            print(command)
            return status

        a, b = 1, 1
        total = 0
        for i in range(self.retries + 1):
            start = perf_counter_ns()
            result = subprocess.run(  # noqa: PLW1510
                command, **self.subprocess_kwargs  # noqa: S603
            )
            end = perf_counter_ns()
            total += end - start
            status = result.returncode
            if status == 0:
                report = lib.CommandReport(
                    command=command, timing_ns=total, retries=i, success=True
                )
                self.reports.append(report)
                return status

            if i != self.retries:
                a, b = b, a + b
                print(f"{command} failed. Retrying in {a}s...")
                sleep(a)

        report = lib.CommandReport(
            command=command, timing_ns=total, retries=i, success=False
        )
        self.reports.append(report)
        return status

    def _parse_recipes(self, parsed_cookbook: dict[str, dict[str, Any]]) -> None:
        for target, raw_recipe in parsed_cookbook.items():
            recipe = lib.Recipe(
                target,
                raw_recipe,
                self.base_dir,
                self.globals.get("vars", {}),
                self.arg_vars,
                extra=[],
                new_order=self._is_new_order,
            )

            if recipe.alias:
                self.aliases[recipe.target] = recipe.alias
            elif recipe.regex:
                self.regex_recipes[recipe.target] = recipe
            else:
                self.static_recipes[recipe.target] = recipe

    def _preprocess_target(self) -> lib.DAG:
        recipe = self._extract_recipe(self.target, use_extra=True)
        if recipe is None:
            raise ValueError(f"No recipe to build {self.target}")

        root = lib.Node(recipe)
        unprocessed = {root.target: root}
        dag = lib.DAG(root)
        while unprocessed and not self.bare:
            target, target_node = unprocessed.popitem()
            dag.add_node(target_node)
            target_recipe = cast(lib.Recipe, target_node.recipe)
            target_recipe.requires.reverse()
            for index, raw_requirement in enumerate(target_recipe.requires):
                recipe = self._extract_recipe(raw_requirement)
                if recipe is None:
                    requirement_path = self._file_path(raw_requirement)
                    requirement = requirement_path.as_posix()
                    if not requirement_path.exists():
                        raise ValueError(f"No recipe to build {requirement}")
                else:
                    requirement = recipe.target
                target_recipe.requires[index] = requirement

                if requirement in dag:
                    node = dag[requirement]
                elif requirement in unprocessed:
                    node = unprocessed[requirement]
                elif recipe is None:
                    node = lib.Node(target=requirement)
                    dag.add_node(node)
                else:
                    node = lib.Node(recipe)
                    unprocessed[requirement] = node
                target_node.add_requirement(node)

        dag.sort()
        self._mark_unchanged(dag)
        if self.verbosity > 3:
            print("=== all targets ===")
            for node in dag:
                print(f"- {node.target}:")
                print(f"    timestamp: {lib.human_readable_timestamp(node.timestamp)}")
                print(f"    should_build: {node.should_build}")
                print(f"    requires: {node.requires}")
                print(f"    required_by: {node.required_by}")
        return dag

    def _make_target(self, recipe: lib.Recipe) -> None:
        if self.verbosity > 1:
            print(f"=== target: {recipe.target} ===")

        for raw_command in recipe.commands:
            command, options = lib.extract_options(raw_command)
            if recipe.echo or "echo" in options or self.verbosity > 2:
                print(command)
            return_code = self._run_command(command)
            if (
                return_code
                and not recipe.allow_failures
                and "allow_failures" not in options
            ):
                if self.print_timing_report:
                    lib.print_reports(self.reports)
                sys.exit(return_code)
        if recipe.phony and recipe.keep_ts:
            path = self._phony_path(recipe.target)
            self.phony_dir.mkdir(exist_ok=True)
            path.touch()
        if recipe.update and not recipe.phony:
            pathlib.Path(recipe.target).touch()

    def _extract_recipe(
        self, target: str, *, use_extra: bool = False
    ) -> lib.Recipe | None:
        if target in self.aliases:
            target = self.aliases[target]

        absolute_path_target = self.base_dir.joinpath(target).as_posix()
        if target in self.static_recipes:
            recipe = self.static_recipes[target]
        elif absolute_path_target in self.static_recipes:
            recipe = self.static_recipes[absolute_path_target]
        else:
            for regex, recipe in self.regex_recipes.items():  # noqa: B007
                if re.fullmatch(regex, target) or re.fullmatch(
                    regex, absolute_path_target
                ):
                    break
            else:
                return None
        extra = self.extra if use_extra else []
        return recipe.for_target(target, extra)

    def _mark_unchanged(self, dag: lib.DAG) -> None:
        for node in dag:
            node.timestamp = self._infer_timestamp(node)
            node.should_build = self._should_build(node)

    def _phony_path(self, target: str) -> pathlib.Path:
        encoded_target = target.replace(".", ".46").replace("/", ".47")
        return self.phony_dir.joinpath(encoded_target)

    def _file_path(self, target: str) -> pathlib.Path:
        return self.base_dir.joinpath(target)

    def _path(self, node: lib.Node) -> pathlib.Path:
        recipe = node.recipe
        if recipe is None or not recipe.phony:
            return self._file_path(node.target)

        return self._phony_path(node.target)

    def _path_exists(self, node: lib.Node) -> bool:
        recipe = node.recipe
        path = self._path(node)
        if recipe is not None and recipe.phony and recipe.existence_command:
            return not self._run_command(recipe.existence_command)

        return path.exists()

    def _should_build(self, node: lib.Node) -> bool:
        recipe = node.recipe
        if recipe is None:
            return False
        if self.force_make:
            return True
        if recipe.phony and recipe.target in self.up_to_date:
            return False
        if not self._path_exists(node):
            return True
        if recipe.exists_only:
            return False

        if not node.requires:
            raise ValueError(
                "This target already exists and has no requirements. "
                "Consider marking it with exists_only"
            )

        if any(child.should_build for child in node.requires):
            return True

        ts = node.timestamp
        req_ts = max(child.timestamp for child in node.requires)
        return req_ts > ts

    def _infer_timestamp(self, node: lib.Node) -> float:
        recipe = node.recipe
        path = self._path(node)
        if recipe is None:
            return path.stat().st_mtime

        if recipe.phony and recipe.target in self.up_to_date:
            return float("inf")

        if not self._path_exists(node):
            return float("inf")

        if not recipe.phony:
            if recipe.recursive:
                descendants = itertools.chain([path], path.rglob("*"))
                return max(p.stat().st_mtime for p in descendants)
            if recipe.exists_only:
                return 0
            return path.stat().st_mtime

        if recipe.exists_only:
            return 0

        if not recipe.keep_ts:
            raise ValueError("Existence commands needs either exists_only or keep_ts")

        return path.stat().st_mtime

    @property
    def _is_new_order(self) -> bool:
        if (raw_order := self.globals.get("new_order")) is None:
            warnings.warn(_msg, lib.RemovedIn60Warning, stacklevel=3)
            return False
        return bool(raw_order)

    def _get_version(self) -> Version:
        if (manual_version := self.globals.get("version")) is None:
            warnings.warn(
                "Min cookbook versions will become mandatory.",
                lib.RemovedIn60Warning,
                stacklevel=3,
            )
            return Version("4.0.0")
        if not isinstance(manual_version, str):
            warnings.warn(
                "Min cookbook versions will be required to be a string",
                lib.RemovedIn60Warning,
                stacklevel=3,
            )
            return Version(str(manual_version))
        return Version(manual_version)


_msg = """ The relative strength of variables will change.
Please set new_order to false in your cookbook to keep the old behaviour,
or to new to adopt the new one.
The new order is the following (from weakest to strongest):
- global variables
- implicit variables (e.g. ${.target})
- regex variables (the ones that come from the target name)
- local variables
- environment variables
- command line arguments
"""



from __future__ import annotations

import math
import os
import re
import shlex
import warnings
from dataclasses import dataclass
from datetime import datetime, timezone
from math import ceil, floor
from pathlib import Path
from typing import Any, Iterator, Literal, Match, cast

from yamk.functions import functions

VAR = re.compile(
    r"(?P<dollars>\$+){(?P<variable>[a-zA-Z0-9_.]+)(?P<sep>:)?(?P<key>[a-zA-Z0-9_.]+)?}"
)
OPTIONS = re.compile(r"\[(?P<options>.*?)\](?P<string>.*)")
FUNCTION = re.compile(r"\$\(\((?P<name>\w+) *(?P<args>.*)\)\)")
SUPPORTED_FILE_EXTENSIONS = {
    ".toml": "toml",
    ".yml": "yaml",
    ".yaml": "yaml",
    ".json": "json",
}


class ANSIEscape:
    __slots__: list[str] = []

    ENDC = "\033[0m"
    BOLD = "\033[1m"
    FAIL = "\033[31m"
    OKGREEN = "\033[32m"
    OKBLUE = "\033[34m"
    WARNING = "\033[33m"


class Recipe:
    def __init__(
        self,
        target: str,
        raw_recipe: dict[str, Any],
        base_dir: Path,
        file_vars: dict[str, Any],
        arg_vars: dict[str, Any],
        extra: list[str],
        original_regex: str | None = None,
        *,
        new_order: bool,
        specified: bool = False,
    ):
        self.new_order = new_order
        self.extra = extra
        self._specified = specified
        self._raw_recipe = raw_recipe
        self.base_dir = base_dir
        self.phony = raw_recipe.get("phony", False)
        self.requires = raw_recipe.get("requires", [])
        self.commands = raw_recipe.get("commands", [])
        self.echo = raw_recipe.get("echo", False)
        self.regex = raw_recipe.get("regex", False)
        self.allow_failures = raw_recipe.get("allow_failures", False)
        self.exists_only = raw_recipe.get("exists_only", False)
        self.keep_ts = raw_recipe.get("keep_ts", False)
        self.existence_command = raw_recipe.get("existence_command", "")
        self.recursive = raw_recipe.get("recursive", False)
        self.update = raw_recipe.get("update", False)
        temp_vars = {
            "global": file_vars,
            "env": dict(**os.environ),
            "arg": arg_vars,
        }
        self.alias = self._alias(raw_recipe.get("alias", False), temp_vars)
        self.target = self._target(target, temp_vars)
        temp_vars["local"] = raw_recipe.get("vars", {})
        if self._specified:
            temp_vars["implicit"] = {
                ".target": self.target,
                ".requirements": self.requires,
                ".extra": self.extra,
            }
            if self.regex:
                if original_regex is None:
                    raise RuntimeError(
                        "original_regex must be specified when target is specific"
                    )
                match_obj = re.fullmatch(original_regex, self.target)
                if match_obj is None:
                    raise RuntimeError(
                        f"original_regex {original_regex} does not match {self.target}"
                    )
                temp_vars["regex"] = match_obj.groupdict()
            else:
                temp_vars["regex"] = {}
        else:
            temp_vars["regex"] = {}
            temp_vars["implicit"] = {}
        self.vars = temp_vars
        if self._specified:
            self._re_evaluate()

    def __str__(self) -> str:
        if self._specified:
            return f"Specified recipe for {self.target}"
        return f"Generic recipe for {self.target}"

    def for_target(self, target: str, extra: list[str]) -> Recipe:
        if self._specified:
            return self
        return self.__class__(
            target,
            self._raw_recipe,
            self.base_dir,
            self.vars["global"],
            self.vars["arg"],
            extra,
            new_order=self.new_order,
            original_regex=self.target,
            specified=True,
        )

    def _evaluate(
        self, obj: Any, variables: dict[str, dict[str, Any]] | None = None
    ) -> Any:
        if variables is None:
            variables = self.vars
        flat_vars = flatten_vars(variables, self.base_dir, new_order=self.new_order)
        parser = Parser(flat_vars, self.base_dir)
        return parser.evaluate(obj)

    def _re_evaluate(self) -> None:
        self.requires = self._evaluate(self.requires)
        self.commands = self._evaluate(self.commands)
        self.existence_command = self._evaluate(self.existence_command)

    def _alias(self, alias: str | Literal[False], variables: dict[str, Any]) -> Any:
        if alias is False:
            return alias
        return self._evaluate(alias, variables)

    def _target(self, target: str, variables: dict[str, Any]) -> Any:
        if not self._specified:
            target = self._evaluate(target, variables)
        if not self.phony and not self.alias:
            target = self.base_dir.joinpath(target).as_posix()
        if self.regex and not self._specified:
            return re.compile(target)
        return target


class Parser:
    def __init__(self, variables: dict[str, Any], base_dir: Path):
        self.vars = variables
        self.base_dir = base_dir

    @staticmethod
    def _stringify(value: Any) -> str:
        if isinstance(value, list):
            return " ".join(map(str, value))
        return str(value)

    def expand_function(self, name: str, args: str) -> Any:
        split_args = shlex.split(args)
        for i, arg in enumerate(split_args):
            split_args[i] = self.evaluate(arg)
        function = functions[name](self.base_dir)
        return function(*split_args)

    def repl(self, match_obj: re.Match[str]) -> str:
        dollars = match_obj.group("dollars")
        variable = match_obj.group("variable")
        key = match_obj.group("key")
        if len(dollars) % 2:
            value = self.vars.get(variable, "")
            if key is None:
                return self._stringify(value)
            if isinstance(value, list):
                key = int(key)
            return self._stringify(value[key])
        return f"{'$'*(len(dollars)//2)}{{{variable}}}"

    def substitute(self, string: str) -> Any:
        function = re.fullmatch(FUNCTION, string)
        if function is not None:
            return self.expand_function(**function.groupdict())

        if (
            string.startswith("$")
            and not string.startswith("$$")
            and re.fullmatch(VAR, string)
        ):
            match = cast(Match[str], re.fullmatch(VAR, string))
            if match["sep"] is None:
                return self.vars[match["variable"]]
        return re.sub(VAR, self.repl, string)

    def evaluate(self, obj: str | list[Any] | dict[str, Any]) -> Any:
        if isinstance(obj, str):
            return self.substitute(obj)
        if isinstance(obj, list):
            out = []
            for string in obj:
                evaluated = self.evaluate(string)
                if isinstance(evaluated, list):
                    out.extend(evaluated)
                else:
                    out.append(evaluated)
            return out
        if isinstance(obj, dict):
            return {
                self.evaluate(key): self.evaluate(value) for key, value in obj.items()
            }
        raise TypeError(f"{obj.__class__.__name__} is not supported for evaluation")


class Node:
    recipe: Recipe | None
    target: str
    timestamp: float
    should_build: bool
    required_by: set[Node]
    requires: list[Node]

    def __init__(self, recipe: Recipe | None = None, *, target: str | None = None):
        self.recipe = recipe
        self.target = target if self.recipe is None else self.recipe.target
        self.requires = []
        self.required_by = set()

    def __str__(self) -> str:
        return self.target

    def __repr__(self) -> str:
        return f"Node <{self}>"

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.target == other.target

    def __hash__(self) -> int:
        return hash(self.target)

    def add_requirement(self, other: Node) -> None:
        if other in self.requires:
            warnings.warn(
                f"`{other}` is included twice in `{self}` requirements, "
                "only the first will be considered",
                RuntimeWarning,
                stacklevel=3,
            )
            return
        self.requires.append(other)
        other.required_by.add(self)


class DAG:
    ordered: list[Node]

    def __init__(self, root: Node):
        self.root = root
        self._mapping = {root.target: root}

    def __getitem__(self, item: str) -> Node:
        return self._mapping[item]

    def __contains__(self, item: str) -> bool:
        return item in self._mapping

    def __iter__(self) -> Iterator[Node]:
        if hasattr(self, "ordered"):
            return iter(self.ordered)
        return iter(self._mapping.values())

    def sort(self) -> None:
        try:
            self.c3_sort()
        except ValueError:
            self.topological_sort()

    def c3_sort(self) -> None:
        self.ordered = self._node_s3_sort(self.root)
        self.ordered.reverse()

    def topological_sort(self) -> None:
        self.ordered = []
        unordered_nodes = self._mapping.copy()
        while unordered_nodes:
            for target, node in unordered_nodes.items():
                if all((parent in self.ordered) for parent in node.requires):
                    del unordered_nodes[target]
                    self.ordered.append(node)
                    break
            else:
                raise ValueError("Cyclic dependencies detected. Cowardly aborting...")
        warnings.warn(
            "The requirements order didn't allow the deterministic order; "
            "fell back to old-style dependency resolution",
            RuntimeWarning,
            stacklevel=3,
        )

    def add_node(self, node: Node) -> None:
        self._mapping[node.target] = node

    def _node_s3_sort(self, node: Node) -> list[Node]:
        out = [node]
        if requirements := node.requires:
            try:
                out.extend(
                    self._merge(
                        *[
                            self._node_s3_sort(requirements)
                            for requirements in requirements
                        ],
                        requirements,
                    )
                )
            except RecursionError as exc:
                raise ValueError("Cannot compute c3_sort") from exc
        return out

    def _merge(self, *node_lists: list[Node]) -> list[Node]:
        result: list[Node] = []
        unmerged = [*node_lists]

        while True:
            if not unmerged:
                return result

            heads = [node_list[0] for node_list in unmerged if node_list]
            tails = [node_list[1:] for node_list in unmerged if len(node_list) > 1]
            for head in heads:
                if head and all(head not in tail for tail in tails):
                    result.append(head)
                    unmerged = self._clean_head(unmerged, head)
                    break
            else:
                raise ValueError("Cannot compute c3_sort")

    @staticmethod
    def _clean_head(unmerged: list[list[Node]], head: Node) -> list[list[Node]]:
        new_list = []
        for original_node_list in unmerged:
            if head == original_node_list[0]:
                if node_list := original_node_list[1:]:
                    new_list.append(node_list)
            else:
                new_list.append(original_node_list)
        return new_list


@dataclass(frozen=True)
class CommandReport:
    command: str
    retries: int
    timing_ns: int
    success: bool

    def _formatted_timing(self) -> str:
        if self.timing_ns < 1000:
            return f"{self.timing_ns} ns"
        timing: float = self.timing_ns
        timing /= 1000
        if timing < 1000:
            return f"{timing:.1f} µs"
        timing /= 1000
        if timing < 1000:
            return f"{timing:.1f} ms"
        timing /= 1000
        return f"{timing:.2f} s"

    def print(self, cols: int) -> None:  # noqa: A003
        if not self.success:
            indicator = "🔴"
            timing_colour = ANSIEscape.FAIL
        elif self.retries:
            indicator = "🟠"
            timing_colour = ANSIEscape.WARNING
        else:
            indicator = "🟢"
            timing_colour = ANSIEscape.OKGREEN
        timing = self._formatted_timing()
        padding = " " * (cols - len(self.command) - len(timing) - 7)
        print(
            indicator,
            f"`{self.command}`",
            padding,
            f"{timing_colour}{timing}{ANSIEscape.ENDC}",
        )


def flatten_vars(
    variables: dict[str, dict[str, Any]], base_dir: Path, *, new_order: bool
) -> dict[str, Any]:
    if new_order:
        order = [
            "env",
            "arg",
            "global",
            "local",
            "global",
            "implicit",
            "regex",
            "local",
            "env",
            "arg",
        ]
    else:
        order = ["env", "global", "regex", "local", "arg", "implicit"]
    output: dict[str, Any] = {}
    strong_keys: set[str] = set()
    for var_type in order:
        var_block = variables.get(var_type, {})
        parser = Parser(output, base_dir)
        for raw_key, raw_value in var_block.items():
            key = parser.evaluate(raw_key)
            key, options = extract_options(key)
            if key in strong_keys:
                continue
            if "weak" in options and key in output:
                continue
            if "strong" in options:
                strong_keys.add(key)
            value = parser.evaluate(raw_value)
            output[key] = value
    return output


def extract_options(string: str) -> tuple[str, set[str]]:
    match = re.fullmatch(OPTIONS, string)
    if match is None:
        return string.strip(), set()

    options = match["options"]
    string = match["string"]
    return string, {s.strip() for s in options.split(",")}


def human_readable_timestamp(timestamp: float) -> str:
    if math.isinf(timestamp):
        return "end of time"
    return str(datetime.fromtimestamp(timestamp, tz=timezone.utc))


def print_reports(reports: list[CommandReport]) -> None:
    cols = os.get_terminal_size().columns
    report_title = " Yam Report "
    padding = "="
    half = (cols - len(report_title)) / 2
    print(
        f"{ANSIEscape.BOLD}{padding * ceil(half)}{report_title}{padding * floor(half)}{ANSIEscape.ENDC}"
    )
    for report in reports:
        report.print(cols)


class RemovedIn60Warning(FutureWarning):
    pass
